#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
a proxomitron helper program
'''
__author__ = 'phoenix (aka whenever)'
__version__ = '0.4b'

from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
from SocketServer import ThreadingMixIn
from urlparse import urlparse, urlunparse
import socket
import ssl

from config import *

class ProxyHandler(BaseHTTPRequestHandler):
    
    server_version = "ProxHTTPSProxy/" + __version__
    # Used to shutdown client socket
    lib_socket = socket.socket
    half_ssl_urls = {}


    def do_METHOD(self):

        # Default settings if no proxy
        (scheme, netloc, path, params, query, fragment) = urlparse(self.path, 'http')
        try:
            remote_host, remote_port = netloc.split(':')
        except ValueError:
            remote_host, remote_port = netloc, '80'
        remote_host_port = '%s:%s' % (remote_host, remote_port)
        host, port = remote_host, remote_port
        request_line = '%s %s %s\r\n' % (self.command, urlunparse(('', '', path, params, query, '')), self.request_version)

        # Overwrite default settings
        if proxy:
            proxy_host, proxy_port = setproxy()
            if 'http' in proxy:
                host, port = proxy_host, proxy_port
            if not half_ssl:
                request_line = '%s %s %s\r\n' % (self.command, self.path, self.request_version)

        server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_sock.connect((host, int(port)))

        visiting_https = True if (remote_host_port in ProxyHandler.half_ssl_urls
                                 or remote_port == '443' and remote_host not in port443_exceptions
                                 or remote_host_port in half_ssl_lists) else False
        if visiting_https:
            if 'http' in proxy:
                https_proxy_init(server_sock, remote_host_port)
            server_sock = ssl.wrap_socket(server_sock)

        headers = {}
        [headers.update({key.lower(): value}) for (key, value) in self.headers.items() if key.lower() not in ('proxy-connection', 'keep-alive')]
        headers['connection'] = 'close'
            
        if 'referer' in headers:
            (scheme, netloc, path, params, query, fragment) = urlparse(headers['referer'], 'http')
            try:
                from_host, from_port = netloc.split(':')
            except ValueError:
                from_host, from_port = netloc, '80'
            from_host_port = '%s:%s' % (from_host, from_port)
            from_https = True if (from_host_port in ProxyHandler.half_ssl_urls
                                 or from_port == '443' and from_host not in port443_exceptions
                                 or from_host_port in half_ssl_lists) else False
            if from_https and visiting_https:
                headers['referer'] = urlunparse(('https', netloc.replace(':443', ''), path, params, query, fragment))
            if from_https and not visiting_https:
                del headers['referer']

        server_sock.sendall(request_line)
        [server_sock.sendall('%s: %s\r\n' % (key, value)) for (key, value) in headers.items()]
        server_sock.sendall('\r\n')

        if self.command == 'POST':
            post_data_len = int(self.headers['Content-Length'])
            post_data = self.rfile.read(post_data_len) if post_data_len > 0 else ''
            if len(post_data) != post_data_len:
                self.send_error(400, 'post data length error')
                return
            server_sock.sendall(post_data)

        # Can't receive response for proxy and non-SSL connection if SHUT_WR here, don't know why
        #if not proxy and half_ssl:
        #    socket.socket.shutdown(server_sock, socket.SHUT_WR)
        response = read_write_response(server_sock, self.wfile)
        server_sock.close()
        
        self.log_message('%s %s %s\r\n', self.command, self.path, self.request_version)
        if debug:
            print '\r\n'.join(sorted(['%s: %s' % (key, value) for (key, value) in headers.items()]))
            if self.command == 'POST':
                print '\r\nPOST data: %s' % post_data
            print '\r\n%s\r\n--------------' % response


    def do_CONNECT(self):
        """Decrypt then encrypt SSL contents with Proxomitron's proxcert.pem
        
        Proxomitron's SSL problem MIGHT arise when servers present newer more
        complicated certificates with extensions that Proxomitron's old SSLEAY
        versions didn't know about.
        
        The problem MIGHT went away with the use of Proxomitron's proxcert.pem,
        which would have been generated simply and does not contain the many
        extensions that have been added to (Open)SSL.
        
        Let Proxomitron use its SSLEAY. Let it manage the in/out HTTP headers
        and filter the data. But defend it from newer incompatible SSL methods.
        
        Ref: http://prxbx.com/forums/showthread.php?tid=1618&pid=14316#pid14316
        """
        
        if proxy:
            proxy_host, proxy_port = setproxy()
        (host, port) = (proxy_host, proxy_port) if 'http' in proxy else self.path.split(':')
        server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        #server_sock = ssl.wrap_socket(server_sock, cert_reqs=ssl.CERT_REQUIRED, ca_certs="/etc/ca_certs_file")
        server_sock.connect((host, int(port)))
        if 'http' in proxy:
            https_proxy_init(server_sock, self.path)
        server_sock = ssl.wrap_socket(server_sock)
        self.wfile.write('%s 200 Connection established\r\n' % self.protocol_version)
        self.wfile.write('Proxy-agent: %s\r\n' % self.version_string())
        self.wfile.write('\r\n')

        client_sock = ssl.wrap_socket(self.connection, certfile=cert_file, server_side=True)

        # Read character-by-character to make sure the whole Initial Request Line
        # was buffered before we do url path extracting
        # Block here until finishing interaction with the SSL warning dialogs
        request_line = ''
        while True:
            chr = client_sock.read(1)
            # EOF before we reach EOL? bad request
            if chr == '':
                client_sock.close()
                return
            request_line += chr
            if chr == '\n':
                break
        method, path, request_version = request_line.split()
        
        # Read headers, and POST data if available
        # Is 0.2s enough to read all data?
        client_sock.settimeout(0.2)
        rawdata = []
        while True:
            try:
                rawdata.append(client_sock.read())
            except ssl.SSLError, e:
                if 'timed out' in str(e).lower():
                    break
                else:
                    raise
            if not rawdata[-1]:
                break
        client_sock.setblocking(True)

        # data is supposed to be ['header1', 'header2', ... , '', 'post_data']
        data = ''.join(rawdata).splitlines()
        headers = {}
        for line in data:
            if line == '':
                break
            if line.startswith((' ','\t')):
                headers[lastname] = '%s %s' % (headers[lastname], line.lstrip())
            name, value = line.split(':', 1)
            headers[name.lower()] = value.strip()
            lastname = name.lower()
        headers['connection'] = 'close'
        # Remove unneeded headers
        [headers.pop(header) for header in headers.keys()  if header in ('keep-alive',)]
        sorted_headers = ''.join(sorted(['%s: %s\r\n' % (name, value) for (name, value) in headers.items()]))
        request = '%s%s\r\n%s' % (request_line, 
                                  sorted_headers,
                                  data[-1])
        server_sock.write(request)
        # Can't receive response for some connection if SHUT_WR here, don't know why
        # Example: https://developer.mozilla.org/
        #if not proxy:
        #    socket.socket.shutdown(server_sock, socket.SHUT_WR)
        response = read_write_response(server_sock, client_sock)
        
        server_sock.close()
        self.lib_socket.shutdown(client_sock, socket.SHUT_WR)
        while True:
            if not client_sock.read():
                break
        client_sock.close()
        
        self.log_message('%s https://%s:%s%s %s\r\n', method, host, port, path, request_version)
        if debug:
            print sorted_headers
            if data[-1]:
                print 'POST data: %s' % data[-1]
            print '\r\n%s\r\n--------------' % response


    def do_HalfSSL_CONNECT(self):
        """Redirect to http location to switch the browser to http mode"""

        host_port = self.path
        self.log_request(307)
        self.wfile.write('%s 200 Connection established\r\n' % self.protocol_version)
        self.wfile.write('Proxy-agent: %s\r\n' % self.version_string())
        self.wfile.write('\r\n')
        ssl_sock = ssl.wrap_socket(self.connection, certfile=cert_file, server_side=True)
        # Read character-by-character to make sure the whole Initial Request Line
        # was buffered before we do url path extracting
        # Block here until finishing interaction with the SSL warning dialogs
        request_line = ''
        while True:
            chr = ssl_sock.read(1)
            # EOF before we reach EOL? bad request
            if chr == '':
                ssl_sock.close()
                return
            if chr == '\n':
                break
            request_line += chr
        location = 'http://%s%s' % (host_port, request_line.split()[1])
        resp = ('HTTP/1.1 307 Moved Temporarily\r\n',
                'Location: %s\r\n' % location,
                'Connection: Close\r\n',
                'Content-Length: 0\r\n',
                '\r\n')
        ssl_sock.write(''.join(resp))
        ProxyHandler.half_ssl_urls[host_port] = True
        
        # a close() with any pending readable data could lead to an immediate reset being sent.
        # http://blog.netherlabs.nl/articles/2009/01/18/the-ultimate-so_linger-page-or-why-is-my-tcp-not-reliable
        #
        # ssl_sock.shutdown() method execute "self._sslobj = None" in line 279 of ssl.py (python 2.6.5)  (bug?)
        # cause any further operation on the socket raising an Exception
        # socket.SHUT_WR == 1
        #ssl_sock.shutdown(1)
        self.lib_socket.shutdown(ssl_sock, socket.SHUT_WR)
        while True:
            if not ssl_sock.read():
                break
        ssl_sock.close()


    do_GET = do_POST = do_HEAD = do_PUT = do_DELETE = do_METHOD
    
    if half_ssl:
        do_CONNECT = do_HalfSSL_CONNECT


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """Handle requests in a separate thread."""
    pass


def setproxy():
    proxy_addr = proxy.split('@')[0]
    proxy_host, proxy_port = proxy_addr.split(':')
    if 'socks5' in proxy:
        import socks
        socks.setdefaultproxy(socks.PROXY_TYPE_SOCKS5, proxy_host, int(proxy_port))
        socket.socket = socks.socksocket
    return proxy_host, proxy_port


def https_proxy_init(proxy_connection, path):

    proxy_connection.sendall('CONNECT %s HTTP/1.0\r\n\r\n' % path)
    # Read out proxy 'connection established' response
    proxy_connection.settimeout(0.1)
    while True:
        try:
            data = proxy_connection.recv(4096)
        except socket.error, e:
            if 'timed out' in str(e).lower():
                break
            else:
                raise
        if not data:
            break
    proxy_connection.setblocking(True)


def read_write_response(server_sock, client_sock):
    """Read response from server_sock and write to client_sock"""
    
    response = ''
    got_all_headers = False
    while True:
        data = server_sock.recv(4096)
        if data:
            client_sock.write(data)
            if not got_all_headers:
                response += data
                if '\r\n\r\n' in response or '\n\n' in response:
                    got_all_headers = True
                    index = response.find('\r\n\r\n') if response.find('\r\n\r\n') != -1 else response.find('\n\n')
                    response = response[:index]
        else:
            break
    return response


if __name__ == '__main__':
    server = ThreadedHTTPServer(('localhost', listen_port), ProxyHandler)
    print '%s serving now, <Ctrl-C> to stop ...' % ProxyHandler.server_version
    print 'Listen Addr  : localhost:%s' % listen_port
    print 'Parent Proxy : %s' % proxy
    print 'Half_SSL     : %s' % (half_ssl and 'Yes' or 'No')
    print 'Debug        : %s' % (debug and 'On' or 'Off')
    print "--------------"
    server.serve_forever()